use std::path::Path;
use std::path::PathBuf;
use std::str::FromStr;
use walkdir::WalkDir;

use func::is_video_file;
pub use video::Video;
mod func;
mod video;

#[derive(Debug, Clone)]
pub struct Library {
    conn: sqlx::PgPool,
}

impl Library {
    pub async fn new() -> Self {
        log::info!("Creating database connection");
        let conn = sqlx::postgres::PgPoolOptions::new()
            .max_connections(5)
            .connect(&std::env::var("DATABASE_URL").unwrap())
            .await
            .unwrap();

        let s = Self { conn };

        sqlx::migrate!("./migrations").run(&s.conn).await.unwrap();

        s
    }
}

// Functions
impl Library {
    // directories

    pub async fn get_directories(&self) -> Vec<String> {
        let res: Vec<(String,)> = sqlx::query_as("SELECT DISTINCT(directory) FROM videos;")
            .fetch_all(&self.conn)
            .await
            .unwrap();
        res.into_iter().map(|x| x.0).collect()
    }

    pub async fn get_directory_videos(&self, dir: &str) -> Vec<Video> {
        sqlx::query_as("SELECT * FROM videos WHERE directory = $1")
            .bind(dir)
            .fetch_all(&self.conn)
            .await
            .unwrap()
    }

    // YT

    pub async fn get_channel_name_yt(&self, id: &str) -> String {
        let name: (String,) =
            sqlx::query_as("SELECT uploader_name FROM youtube_meta WHERE uploader_id = $1")
                .bind(id)
                .fetch_one(&self.conn)
                .await
                .unwrap();

        name.0
    }

    pub async fn get_tags_yt(&self) -> Vec<String> {
        let res: Vec<(String,)> = sqlx::query_as("SELECT DISTINCT(tag) FROM youtube_meta_tags")
            .fetch_all(&self.conn)
            .await
            .unwrap();
        res.into_iter().map(|x| x.0).collect()
    }

    pub async fn get_videos_by_tag_yt(&self, tag: &str) -> Vec<Video> {
        sqlx::query_as(
            "SELECT * FROM youtube_meta_tags INNER JOIN youtube_meta ON youtube_meta_tags.youtube_id = youtube_meta.id  INNER JOIN videos ON videos.youtube_id = youtube_meta.id WHERE tag = $1")
            .bind(tag)
            .fetch_all(&self.conn).await.unwrap()
    }

    pub async fn get_channel_videos_yt(&self, id: &str) -> Vec<Video> {
        sqlx::query_as(
            "SELECT * FROM youtube_meta INNER JOIN videos ON youtube_meta.id = videos.youtube_id WHERE uploader_id = $1 ORDER BY youtube_meta.upload_date DESC;")
            .bind(id)
            .fetch_all(&self.conn).await.unwrap()
    }

    // videos

    pub async fn get_random_videos(&self, n: i64) -> Vec<Video> {
        sqlx::query_as("SELECT * FROM videos ORDER BY RANDOM() LIMIT $1;")
            .bind(n)
            .fetch_all(&self.conn)
            .await
            .unwrap()
    }

    pub async fn get_video_by_id(&self, id: &str) -> Option<Video> {
        sqlx::query_as("SELECT * FROM videos WHERE id = $1")
            .bind(uuid::Uuid::from_str(id).unwrap_or(uuid::Uuid::nil()))
            .fetch_optional(&self.conn)
            .await
            .unwrap()
    }

    pub async fn get_video_by_youtube_id(&self, id: &str) -> Option<Video> {
        sqlx::query_as("SELECT * FROM videos WHERE youtube_id = $1")
            .bind(id)
            .fetch_optional(&self.conn)
            .await
            .unwrap()
    }

    pub async fn has_path(&self, path: &Path) -> bool {
        sqlx::query("SELECT path FROM videos WHERE path = $1")
            .bind(path.display().to_string())
            .fetch_optional(&self.conn)
            .await
            .unwrap()
            .is_some()
    }

    // search

    pub async fn search_video(&self, query: &str, start: i64, n: i64) -> Vec<Video> {
        let query = format!("%{query}%");

        let ids: Vec<(uuid::Uuid,)> = sqlx::query_as(
            r#"SELECT DISTINCT
            vm.id,
            ( -- Calculate a score for the video based on matches
                (ym.title LIKE $1) +
                (ym.description LIKE $1) +
                (ym.uploader_name LIKE $1) +
                (vm.directory LIKE $1)
            ) AS score
        FROM
            youtube_meta AS ym
        LEFT JOIN
            videos AS vm ON ym.id = vm.youtube_id
        LEFT JOIN
            youtube_meta_tags AS ymt ON ym.id = ymt.youtube_id
        WHERE
            (ym.title LIKE $1) OR
            (ym.description LIKE $1) OR
            (ym.uploader_name LIKE $1) OR
            (vm.directory LIKE $1) OR
            (ymt.tag LIKE $1)
        ORDER BY
            score DESC,
            ym.upload_date DESC LIMIT $2 OFFSET $3;"#,
        )
        .bind(query)
        .bind(n)
        .bind(start)
        .fetch_all(&self.conn)
        .await
        .unwrap();

        let mut videos = Vec::new();

        for id in ids {
            videos.push(Video::get(&id.0).await.unwrap());
        }

        videos
    }
}

// video library scan
impl Library {
    pub async fn scan_dir(&self, dir: &PathBuf) {
        log::info!("Scanning {dir:?}");
        let lib = self.get_video_paths(dir).await;

        for path in lib {
            if !Video::has_path(&path.display().to_string()).await {
                Video::insert_path_to_db(&path).await;
            }
        }

        log::info!("Finished scanning {dir:?}");
    }

    async fn get_video_paths(&self, dir: &PathBuf) -> Vec<PathBuf> {
        let mut videos: Vec<PathBuf> = vec![];

        for entry in WalkDir::new(dir).follow_links(true) {
            match entry {
                Ok(entry) => {
                    if entry.file_type().is_file() {
                        log::info!("Scanning {entry:?}");
                        let file_name = entry.file_name().to_string_lossy();
                        if is_video_file(&file_name) {
                            let video_path = entry.path().to_path_buf();

                            if self.has_path(&video_path).await {
                                continue;
                            }

                            videos.push(video_path);
                        }
                    }
                }
                Err(err) => eprintln!("Error reading entry: {err}"),
            }
        }
        videos
    }
}
