use core::num;

use maud::{html, PreEscaped};

use crate::library::Video;

use rocket::{
    http::{ContentType, Status},
    request::{self, FromRequest, Request},
};

pub struct HTMX {
    is_htmx: bool,
}

#[rocket::async_trait]
impl<'r> FromRequest<'r> for HTMX {
    type Error = ();

    async fn from_request(req: &'r Request<'_>) -> request::Outcome<Self, Self::Error> {
        rocket::outcome::Outcome::Success(HTMX {
            is_htmx: !req
                .headers()
                .get("HX-Request")
                .collect::<Vec<&str>>()
                .is_empty(),
        })
    }
}

pub fn htmx_link(
    url: &str,
    class: &str,
    onclick: &str,
    content: PreEscaped<String>,
) -> PreEscaped<String> {
    html!(
        a class=(class) onclick=(onclick) href=(url) hx-get=(url) hx-target="#main_content" hx-push-url="true" hx-swap="innerHTML" {
            (content);
        };
    )
}

pub fn script(script: &str) -> PreEscaped<String> {
    html!(
        script {
            (PreEscaped(script))
        };
    )
}

pub fn shell(content: PreEscaped<String>, title: &str) -> PreEscaped<String> {
    html! {
        html {
            head {
                title { (title) };
                script src="https://cdn.tailwindcss.com" {};
                script src="https://unpkg.com/htmx.org@2.0.3" integrity="sha384-0895/pl2MU10Hqc6jd4RvrthNlDiE9U1tWmX7WRESftEDRosgxNsQG/Ze9YMRzHq" crossorigin="anonymous" {};
                meta name="viewport" content="width=device-width, initial-scale=1.0";
            };
            body class="bg-black text-white" {
                header class="bg-gray-800 text-white shadow-md py-2" {
                    (script(include_str!("../scripts/header.js")));

                    div class="flex justify-start px-6" {

                        (htmx_link("/", "flex items-center space-x-2", "stopAllVideos()", html!(
                            img src="/favicon" alt="Logo" class="w-10 h-10 rounded-md";
                            span class="font-semibold text-xl" { "WatchDogs" };
                        )))

                        };

                    };
                };

                div id="main_content" {
                    (content)
                };
            };
    }
}

pub fn render_page(
    htmx: HTMX,
    content: PreEscaped<String>,
    title: &str,
) -> (Status, (ContentType, String)) {
    if !htmx.is_htmx {
        (
            Status::Ok,
            (ContentType::HTML, shell(content, title).into_string()),
        )
    } else {
        (Status::Ok, (ContentType::HTML, content.into_string()))
    }
}

pub fn loading_spinner() -> PreEscaped<String> {
    html! {
        style {
            ".spinner { display: flex;justify-content: center;align-items: center;height: 100vh;}
            .spinner-border { border: 2px solid #007bff;border-top: 2px solid transparent;border-radius: 50%;width: 40px;height: 40px;animation: spin 1s linear infinite;}
            @keyframes spin {0% { transform: rotate(0deg); }100% { transform: rotate(360deg); }}"
        };
        div class="spinner" {
            div class="spinner-border" {};
        };
    }
}

pub fn search_bar(query: &str) -> PreEscaped<String> {
    html! {
        form hx-get="/search" action="/search" hx-push-url="true" hx-target="#main-view" hx-swap="innerHTML" {
            input style="width: 100%;" value=(query) name="query" type="search" placeholder="Search...";
        };
    }
}

fn format_seconds_to_hhmmss(seconds: f64) -> String {
    let total_seconds = seconds as u64;
    let hours = total_seconds / 3600;
    let minutes = (total_seconds % 3600) / 60;
    let seconds = total_seconds % 60;
    if hours != 0 {
        format!("{:02}:{:02}:{:02}", hours, minutes, seconds)
    } else {
        format!("{:02}:{:02}", minutes, seconds)
    }
}

pub fn format_date(date: &chrono::NaiveDate) -> String {
    // TODO : Implement
    date.to_string()
}

pub fn format_number(num: i32) -> String {
    // TODO : Implement
    num.to_string()
}

pub async fn video_element_wide(video: &mut Video) -> PreEscaped<String> {
    htmx_link(
        &format!("/watch?v={}", video.id),
        "flex items-center w-full p-4 bg-gray-900 shadow-lg rounded-lg overflow-hidden mb-2 mt-2",
        "stopAllVideos()",
        html!(
                div class="flex-shrink-0" {
                    img width="480" src=(format!("/video/thumbnail?v={}", video.id)) alt="Video Thumbnail" class="w-48 h-32 object-cover rounded-md";
                };

                div class="flex flex-col flex-grow ml-4" {
                    h3 class="text-lg font-semibold truncate mb-1" {
                        ( video.title )
                    };

                    @if let Some(meta) = video.youtube_meta().await {
                    div class="text-sm text-gray-400 mb-2" {
                        span class="font-medium" { ( meta.uploader_name ) }
                        span { " - " }
                        span { ( format_date(&meta.upload_date) ) }
                    };

                    div class="text-sm text-gray-400" {
                        span { ( format_number(meta.views) ) }
                        span { " views" }
                        span { " - " }
                        span class="ml-2 bg-black text-white text-xs px-2 py-1 rounded-sm opacity-90" {
                            (( format_seconds_to_hhmmss(video.duration) ))
                        };
                    };
                };
            };
        ),
    )
}

pub async fn video_element(video: &mut Video) -> PreEscaped<String> {
    htmx_link(
        &format!("/watch?v={}", video.id),
        "max-w-sm mx-auto p-4 max-h-60 aspect-video",
        "stopAllVideos()",
        html!(
                div class="bg-gray-900 shadow-lg rounded-lg overflow-hidden" {
                    div class="relative" {
                        img width="480" src=(format!("/video/thumbnail?v={}", video.id)) alt="Video Thumbnail" class="w-full h-auto object-cover aspect-video";
                        span class="absolute bottom-2 right-2 bg-black text-white text-xs px-2 py-1 rounded-sm opacity-90" {
                            (( format_seconds_to_hhmmss(video.duration) ))
                        };
                    };

                    div class="bg-gray-900 shadow-lg rounded-lg overflow-hidden" {
                        div class="p-4" {
                            h3 class="text-lg font-semibold truncate" {
                                ( video.title )
                            };
                        };
                    };
                };
        ),
    )
}
